import logging

from sasktran import LineOfSight, Geometry, Geodetic, StokesVector
from skretrieval.core.sensor import Sensor, OpticalGeometry
from skretrieval.core.lineshape import LineShape, Gaussian
from skretrieval.core.sensor.spectrograph import SpectrographFast, Spectrograph
from skretrieval.core.radianceformat import RadianceSpectralImage
from typing import List, Type, Union
import xarray as xr
import numpy as np


class SpectralImager(Sensor):
    """
    Basic ideal instrument for spectral images on a rectangular 2d grid.
    Contains an array of SpectrographFast instances to do the calculation for each pixel.

    Parameters
    ----------
    wavelength_nm : np.ndarray
        Array of central wavelengths that the sensor measures.
    spectral_lineshape : LineShape
        Spectral point spread function of the detector. Each pixel is assumed to have the same PSF
    pixel_vert_fov : LineShape
        Vertical angular point spread function of each pixel. Each pixel is assumed to have the same PSF
    pixel_horiz_fov : LineShape
        Horizontal angular point spread function of each pixel. Each pixel is assumed to have the same PSF
    image_horiz_fov : float
        horizontal field of view of the entire detector in degrees
    image_vert_fov : float
        vertical field of view of the entire detector in degrees
    num_columns: int
        Number of columns in the sensor
    num_rows : int
        Number of rows in the sensor
    """
    def __init__(self, wavelength_nm: np.ndarray,
                 spectral_lineshape: Union[LineShape, List[LineShape]],
                 pixel_vert_fov: LineShape, pixel_horiz_fov: LineShape,
                 image_horiz_fov: float, image_vert_fov: float,
                 num_columns: int, num_rows: int):

        self.horizontal_fov = image_horiz_fov
        self.vertical_fov = image_vert_fov
        self._num_columns = num_columns
        self._num_rows = num_rows
        self._wavelength_nm = wavelength_nm
        self._horiz_fov = pixel_horiz_fov
        self._vert_fov = pixel_vert_fov
        self._pixels = self._create_pixels(wavelength_nm, spectral_lineshape, pixel_vert_fov, pixel_horiz_fov)

    @property
    def num_columns(self) -> int:
        return self._num_columns

    @property
    def num_rows(self) -> int:
        return self._num_rows

    @property
    def pixels(self):
        return self._pixels

    def optical_geometries(self, optical_geometry, num_columns: int = None, num_rows: int = None):

        if num_columns is None:
            num_columns = self.num_columns

        if num_rows is None:
            num_rows = self.num_rows

        return self.pixel_optical_axes(optical_geometry,
                                       self.horizontal_fov, self.vertical_fov,
                                       num_columns, num_rows)

    def _create_pixels(self, wavelength_nm: np.ndarray,
                       spectral_lineshape: LineShape,
                       pixel_vert_fov: LineShape,
                       pixel_horiz_fov: LineShape) -> List[Spectrograph]:
        pixels = []

        for i in range(self._num_columns):
            for j in range(self._num_rows):
                pixels.append(SpectrographFast(wavelength_nm=wavelength_nm, pixel_shape=spectral_lineshape,
                                               vert_fov=pixel_vert_fov, horiz_fov=pixel_horiz_fov))

        return pixels

    def model_radiance(self, optical_geometry: OpticalGeometry, model_wavel_nm: np.array, model_geometry: Geometry,
                       radiance: np.array, wf=None) -> RadianceSpectralImage:

        optical_axes = self.optical_geometries(optical_geometry)
        model_values = [p.model_radiance(oa, model_wavel_nm, model_geometry, radiance, wf=wf)
                        for p, oa in zip(self._pixels, optical_axes)]

        return RadianceSpectralImage(xr.concat([m.data for m in model_values], dim='los'),
                                     num_columns=self._num_columns)

    def radiance_format(self) -> Type[RadianceSpectralImage]:

        return RadianceSpectralImage

    def measurement_geometry(self, optical_geometry: OpticalGeometry,
                             num_columns: int = None,
                             num_rows: int = None) -> List[LineOfSight]:

        if num_rows is None:
            num_rows = self._num_rows
        if num_columns is None:
            num_columns = self._num_columns

        # samples = Sampling(self, optical_geometry)
        optical_axes = self.optical_geometries(optical_geometry, num_columns=num_columns, num_rows=num_rows)

        return [LineOfSight(mjd=oa.mjd, observer=oa.observer, look_vector=oa.look_vector)
                for oa in optical_axes]

    def measurement_wavelengths(self) -> np.ndarray:
        return np.unique(np.array([p.measurement_wavelengths() for p in self._pixels]))

    def required_wavelengths(self, res_nm: float) -> np.ndarray:
        return np.unique(np.array([p.required_wavelengths(res_nm) for p in self._pixels]))

    @staticmethod
    def project_to_plane(optical_axis: OpticalGeometry, model_geometry: Union[List[OpticalGeometry], Geometry]):
        """
        Return the vertical and horizontal angles in radians of the `model_geometry` with respect to the `optical_axis`.

        Parameters
        ----------
        optical_axis
        model_geometry
        """
        x_axis = np.array(optical_axis.look_vector)
        vert_normal = np.cross(np.array(x_axis), np.array(optical_axis.local_up))
        vert_normal = vert_normal / np.linalg.norm(vert_normal)
        vert_y_axis = np.cross(vert_normal, x_axis)

        horiz_y_axis = vert_normal

        horiz_angle = []
        vert_angle = []

        try:
            lines_of_sight = model_geometry.lines_of_sight
        except AttributeError:
            lines_of_sight = model_geometry

        for los in lines_of_sight:
            vert_angle.append(np.arctan2(np.dot(los.look_vector, vert_y_axis), np.dot(los.look_vector, x_axis)))
            horiz_angle.append(np.arctan2(np.dot(los.look_vector, horiz_y_axis), np.dot(los.look_vector, x_axis)))

        return np.array(horiz_angle), np.array(vert_angle)

    def project_to_detector(self, optical_axis: OpticalGeometry,
                            model_geometry: Union[List[OpticalGeometry], Geometry]):
        """
        Return the vertical and horizontal positions of the `model_geometry`  on the detector plane in units of pixels.
        """

        hfov = self.horizontal_fov * np.pi / 180
        vfov = self.vertical_fov * np.pi / 180

        hangle, vangle = self.project_to_plane(optical_axis, model_geometry)
        hpixel = (hangle + hfov / 2) * self.num_columns
        vpixel = (vangle + vfov / 2) * self.num_rows

        return hpixel, vpixel

    @staticmethod
    def pixel_optical_axes(optical_axis: OpticalGeometry,
                           hfov: float, vfov: float,
                           num_columns: int, num_rows: int) -> List[OpticalGeometry]:
        """
        Get the optical geometry at the center of each pixel in the sensor.
        
        Parameters
        ----------
        optical_axis : OpticalGeometry
            The optical axis of the center of the sensor
        hfov : float
            horizontal field of view in degrees
        vfov : float
            vertical field of view in degrees
        num_columns :
            Number of colums in the sensor
        num_rows :
            Number of rows in the sensor

        Returns
        -------
        List[OpticalGeometry]
            The optical geoemetry of each pixel in the sensor as a row-major list.
        """
        pixel_geometry = []
        look = optical_axis.look_vector
        up = optical_axis.local_up
        mjd = optical_axis.mjd
        obs = optical_axis.observer

        dist = 1e6  # used for scaling purposes to improve numerical precision
        htan = np.tan(hfov / 2 * np.pi / 180) * dist
        vtan = np.tan(vfov / 2 * np.pi / 180) * dist

        if num_columns == 1:
            x = np.array([0.0])
        else:
            x = np.linspace(-htan, htan, num_columns)

        if num_rows == 1:
            y = np.array([0.0])
        else:
            y = np.linspace(-vtan, vtan, num_rows)

        nx, ny = np.meshgrid(x, y)
        center = obs + look * dist
        xaxis = np.cross(up, look)
        xaxis /= np.linalg.norm(xaxis)
        yaxis = up

        nyf = ny.flatten()
        nxf = nx.flatten()
        pos = center + xaxis * nxf[:, np.newaxis] + yaxis * nyf[:, np.newaxis]
        los = pos - obs
        los = los / np.linalg.norm(los, axis=1)[:, np.newaxis]
        pixel_yaxis = np.cross(los, up, axisa=1)
        pixel_yaxis = pixel_yaxis / np.linalg.norm(pixel_yaxis, axis=1)[:, np.newaxis]
        pixel_up = np.cross(pixel_yaxis, los, axisa=1, axisb=1)
        pixel_up = pixel_up / np.linalg.norm(pixel_up, axis=1)[:, np.newaxis]

        # TODO: the list allocation is kind of slow, can this be improved?
        pixel_geometry = [OpticalGeometry(observer=obs, look_vector=look, local_up=u, mjd=mjd)
                          for look, u in zip(los, pixel_up)]

        # for x, y, in zip(nx.flatten(), ny.flatten()):
        #     pos = center + xaxis * x + yaxis * y
        #     los = pos - obs
        #     los /= np.linalg.norm(los)
        #     pixel_yaxis = np.cross(los, up)
        #     pixel_yaxis /= np.linalg.norm(pixel_yaxis)
        #     pixel_up = np.cross(pixel_yaxis, los)
        #     pixel_up /= np.linalg.norm(pixel_up)
        #     pixel_geometry.append(OpticalGeometry(observer=obs, look_vector=los, local_up=pixel_up, mjd=mjd))

        return pixel_geometry


class SpectralImagerFast(SpectralImager):
    """
    Similar to `SpectralImager`, but cannot have a variable `LineShape` for different pixels.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._los_interpolator = None

    def model_radiance(self, optical_geometry: OpticalGeometry, model_wavel_nm: np.array, model_geometry: Geometry,
                       radiance: np.array, wf=None) -> RadianceSpectralImage:

        optical_axes = self.optical_geometries(optical_geometry)
        wavel_interp, _ = self._pixels[0]._construct_interpolators(model_geometry, optical_axes[0], model_wavel_nm)

        # if self._los_interpolator is None:
        self._los_interpolator = self._construct_los_interpolator(model_geometry, optical_axes)

        if type(radiance) is xr.Dataset:
            if 'I' in radiance.keys():
                radiance = radiance.I.values
        elif type(radiance[0][0]) is StokesVector:
            radiance = np.array([[r.I for r in rad] for rad in radiance])

        modelled_radiance = np.einsum('ij,jk...,kl', wavel_interp, radiance, self._los_interpolator, optimize='optimal')
        data = xr.Dataset({'radiance': (['wavelength', 'los'], modelled_radiance),
                           'mjd': (['los'], np.array([oa.mjd for oa in optical_axes])),
                           'los_vectors': (['los', 'xyz'], np.array([oa.look_vector for oa in optical_axes])),
                           'observer_position': (['los', 'xyz'], np.array([oa.observer for oa in optical_axes]))},
                          coords={'wavelength': self.measurement_wavelengths(),
                                  'xyz': ['x', 'y', 'z']})

        if wf is not None:
            if type(wf) is xr.Dataset:
                for key in wf:
                    modelled_wf = np.einsum('ij,jkl,km->iml', wavel_interp, wf[key], self._los_interpolator, optimize='optimal')
                    data[key] = (wf[key].dims, modelled_wf)
            else:
                modelled_wf = np.einsum('ij,jkl,km->iml', wavel_interp, wf, self._los_interpolator, optimize='optimal')
                data['wf'] = (wf.dims, modelled_wf)

        return RadianceSpectralImage(data, num_columns=self._num_columns)

    def _construct_los_interpolator(self, model_geometry, optical_axes: List[OpticalGeometry]):
        """
        Internally constructs the matrix used for the line of sight interpolation
        """
        x_axis = np.array([oa.look_vector for oa in optical_axes])
        local_up = np.array([oa.local_up for oa in optical_axes])
        vert_normal = np.cross(x_axis, local_up, axisa=1, axisb=1)
        vert_normal = vert_normal / np.linalg.norm(vert_normal, axis=1)[:, np.newaxis]
        vert_y_axis = np.cross(vert_normal, x_axis, axisa=1, axisb=1)

        horiz_y_axis = vert_normal
        geom_los = np.array([los.look_vector for los in model_geometry.lines_of_sight])

        los_dot_x = np.einsum('ij,kj->ik', geom_los, x_axis)
        vert_angle = np.arctan2(np.einsum('ij,kj->ik', geom_los, vert_y_axis), los_dot_x)
        horiz_angle = np.arctan2(np.einsum('ij,kj->ik', geom_los, horiz_y_axis), los_dot_x)

        horiz_interpolator = np.ones_like(horiz_angle)
        vert_interpolator = np.ones_like(horiz_angle)
        for idx in range(len(optical_axes)):
            horiz_interpolator[:, idx] = self._horiz_fov.integration_weights(0, horiz_angle[:, idx])
            vert_interpolator[:, idx] = self._vert_fov.integration_weights(0, vert_angle[:, idx])

        # TODO: Only valid for exponential distributions?
        los_interpolator = horiz_interpolator * vert_interpolator
        los_interpolator = los_interpolator / np.nansum(los_interpolator, axis=0)[np.newaxis, :]

        if np.any(np.isnan(los_interpolator)):
            logging.warning('NaN encountered in SpectralImagerFast los constructor')

        return los_interpolator
